#include <Python.h>

#include <XSsymbol.h>
#include <XSstreams.h>
#include <XSContainer.h>
#include <xsTypes.h>
#include <XSFit/Fit/Fit.h>
#include <XSFunctions/functionMap.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/GlobalContainer/ResponseContainer.h>
#include <XSUser/Global/Global.h>
#include <XSUser/Global/XSGlobal.h>
#include <XSUser/Python/PyFit.h>
#include <XSUser/Python/PyModel.h>
#include <XSUser/Python/PyPlot.h>
#include <XSUser/Python/PySpectrum.h>
#include <XSUser/UserInterface/TclStream.h>
#include <XSUtil/FunctionUtils/XSModelFunction.h>
#include <XSUtil/Numerics/RandomGenerator.h>
#include <XSUtil/Utils/IosHolder.h>
#include <XSUtil/Utils/XSstream.h>

void xspecLibInit();

PyMethodDef _PyXspecMethods[] = {
   {"doFit",_pyXspec_doFit, METH_VARARGS,"Perform a fit using currently loaded data and models."},
   
   {"createModel",_pyXspec_createModel, METH_VARARGS,"Create a new model."},
   
   {"readSpectrum",_pyXspec_readSpectrum, METH_VARARGS,"Read a spectral data file."},
   
   {"setDevice",_pyXspec_setDevice, METH_VARARGS,"Set the plotting device."},
   {"doPlot",_pyXspec_doPlot, METH_VARARGS,"Display the plot."},
   {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC
init_pyXspec(void)
{
   (void) Py_InitModule("_pyXspec", _PyXspecMethods);
   // Anything that throws from here should cause immediate program termination.
   xspecLibInit();
}

void xspecLibInit()
{
   using namespace XSGlobal;
   
   globalData = GlobalData::Instance();
   globalData->readSettings();
   
   // If we write a new XSchannel subclass, won't need to use TclIO here.
   // (interp must still be set due to usage in GlobalData::processSettings,
   // called by XSGlobal::startUp().)
   interp = Tcl_CreateInterp();
   Tcl_Init(interp);
   XSstream::defineChannel(tcin, new TclIO(Tcl_GetStdChannel(TCL_STDIN)), -32768);
   tcin.tie(&tcout);
   XSstream::defineChannel(tcout, new TclIO(Tcl_GetStdChannel(TCL_STDOUT)), bufSize);
   XSstream::defineChannel(tcerr, new TclIO(Tcl_GetStdChannel(TCL_STDERR)), 0);
   IosHolder::setStreams(&tcin, &tcout, &tcerr);
   
   createFunctionMap();

// create model & response containers. This will eventually be in the equivalent call to xs_start
// will also need to create data container.
 

   Parameter::initializeLinks();

   string modelDat(globalData->managerDir());
   modelDat += "/";
   modelDat += globalData->modDescripFile();

   XSModelFunction::updateComponentList(modelDat, true);

   registerFunctionUtility();

   DummyResponse* defaultEnergyArrayResponse (globalData->getDummyResponse());

   XSContainer::xsRegistry = XspecRegistry::Instance();

   XSContainer::datasets  = XSContainer::DataContainer::Instance();

   XSContainer::models    = XSContainer::ModelContainer::Instance();

   XSContainer::responses = XSContainer::ResponseContainer::Instance(defaultEnergyArrayResponse);

   // Provide a default initial seed to RandomGenerator singleton based on clock time.
   Numerics::DefaultRandomGenerator& randGen = 
           Numerics::DefaultRandomGenerator::instance();
   randGen.seed(time(0));
   randGen.initialize();

   // need to register Fit, Statistic and Weightings so we can check  
   // whether settings file makes any sense.
   registerNativeFitMethods();

   // also registers Statistical Weighting methods.
   registerNativeStatMethods();

   //registers all known table formats
   registerNativeTables();        

   // This is deliberately not a reference.  Want to store this 
   // even after globalData's method may be changed by init file.
   const string defaultMethod = globalData->method();      
   XSContainer::fit = Fit::Instance(defaultMethod,globalData->statistic());

   registerNativeRandomizingStrategies(XSContainer::fit);

   registerNativeLineLists();

   // initialize user settings and interface
   // Perhaps startUp should be modified so that it does NOT report 
   // a version string.  Till then we need to do this:
   string binaryLoc(getenv("HEADAS"));
   binaryLoc += "/bin/xspec";
   static string XSPEC_CURRENT("XSPEC_CURRENT=");
   XSPEC_CURRENT += binaryLoc;
   putenv(const_cast<char*>(XSPEC_CURRENT.c_str()));
      
   XSGlobal::startUp();
   
   return;
}
